from __future__ import annotations

from ._loader import ActLoad as ActLoad
from ._saver import Activation as Activation
from ._saver import ActSave as ActSave
from ._saver import Transform as Transform



from __future__ import annotations

import pprint
from dataclasses import dataclass, field
from functools import cached_property
from logging import getLogger
from pathlib import Path
from typing import cast, overload

import numpy as np
from typing_extensions import TypeVar, override

log = getLogger(__name__)

T = TypeVar("T", infer_variance=True)


@dataclass
class LoadedActivation:
    base_dir: Path = field(repr=False)
    name: str
    num_activations: int = field(init=False)
    activation_files: list[Path] = field(init=False, repr=False)

    def __post_init__(self):
        if not self.activation_dir.exists():
            raise ValueError(f"Activation dir {self.activation_dir} does not exist")

        # The number of activations = the * of .npy files in the activation dir
        self.activation_files = list(self.activation_dir.glob("*.npy"))
        # Sort the activation files by the numerical index in the filename
        self.activation_files.sort(key=lambda p: int(p.stem))
        self.num_activations = len(self.activation_files)

    @property
    def activation_dir(self) -> Path:
        return self.base_dir / self.name

    def _load_activation(self, item: int):
        activation_path = self.activation_files[item]
        if not activation_path.exists():
            raise ValueError(f"Activation {activation_path} does not exist")
        return cast(np.ndarray, np.load(activation_path, allow_pickle=True))

    @overload
    def __getitem__(self, item: int) -> np.ndarray: ...

    @overload
    def __getitem__(self, item: slice | list[int]) -> list[np.ndarray]: ...

    def __getitem__(
        self, item: int | slice | list[int]
    ) -> np.ndarray | list[np.ndarray]:
        if isinstance(item, int):
            return self._load_activation(item)
        elif isinstance(item, slice):
            return [
                self._load_activation(i)
                for i in range(*item.indices(self.num_activations))
            ]
        elif isinstance(item, list):
            return [self._load_activation(i) for i in item]
        else:
            raise TypeError(f"Invalid type {type(item)} for item {item}")

    def __iter__(self):
        return iter(self[i] for i in range(self.num_activations))

    def __len__(self):
        return self.num_activations

    def all_activations(self):
        return [self[i] for i in range(self.num_activations)]

    @override
    def __repr__(self):
        return f"<LoadedActivation {self.name} ({self.num_activations} activations)>"


class ActLoad:
    @classmethod
    def all_versions(cls, dir: str | Path):
        dir = Path(dir)

        # If the dir is not an activation base directory, we return None
        if not (dir / ".activationbase").exists():
            return None

        # The contents of `dir` should be directories, each of which is a version.
        return [
            (subdir, int(subdir.name)) for subdir in dir.iterdir() if subdir.is_dir()
        ]

    @classmethod
    def is_valid_activation_base(cls, dir: str | Path):
        return cls.all_versions(dir) is not None

    @classmethod
    def from_latest_version(cls, dir: str | Path):
        # The contents of `dir` should be directories, each of which is a version
        # We need to find the latest version
        if (all_versions := cls.all_versions(dir)) is None:
            raise ValueError(f"{dir} is not an activation base directory")

        path, _ = max(all_versions, key=lambda p: p[1])
        return cls(path)

    def __init__(
        self,
        dir: Path | None = None,
        *,
        _base_activations: dict[str, LoadedActivation] | None = None,
        _prefix_chain: list[str] | None = None,
    ):
        """Initialize ActLoad from a directory or from filtered activations.

        Args:
            dir: Path to the activation directory. Required for root ActLoad instances.
            _base_activations: Pre-filtered activations dict. Used internally for prefix filtering.
            _prefix_chain: Chain of prefixes that have been applied. Used for repr.
        """
        self._dir = dir
        self._base_activations = _base_activations
        self._prefix_chain = _prefix_chain or []

    def activation(self, name: str):
        if self._dir is None:
            raise ValueError(
                "Cannot create activation from filtered ActLoad without base directory"
            )
        # For filtered instances, we need to reconstruct the full name
        full_name = "".join(self._prefix_chain) + name
        return LoadedActivation(self._dir, full_name)

    @cached_property
    def activations(self):
        if self._base_activations is not None:
            return self._base_activations

        if self._dir is None:
            raise ValueError("ActLoad requires either dir or _base_activations")

        dirs = list(self._dir.iterdir())
        # Sort the dirs by the last modified time
        dirs.sort(key=lambda p: p.stat().st_mtime)

        return {p.name: LoadedActivation(self._dir, p.name) for p in dirs}

    def __iter__(self):
        return iter(self.activations.values())

    def __getitem__(self, item: str):
        return self.activations[item]

    def __len__(self):
        return len(self.activations)

    def _ipython_key_completions_(self):
        return list(self.activations.keys())

    @override
    def __repr__(self):
        acts_str = pprint.pformat(
            {
                name: f"<{activation.num_activations} activations>"
                for name, activation in self.activations.items()
            }
        )
        acts_str = acts_str.replace("'<", "<").replace(">'", ">")

        if self._prefix_chain:
            prefix_str = "".join(self._prefix_chain)
            return f"ActLoad(prefix='{prefix_str}', {acts_str})"
        else:
            return f"ActLoad({acts_str})"

    def get(self, name: str, /, default: T) -> LoadedActivation | T:
        return self.activations.get(name, default)

    def filter_by_prefix(self, prefix: str) -> ActLoad:
        """Create a filtered view of activations that match the given prefix.

        Args:
            prefix: The prefix to filter by. Only activations whose names start
                   with this prefix will be included in the filtered view.

        Returns:
            A new ActLoad instance that provides access to matching activations
            with the prefix stripped from their keys. Can be chained multiple times.

        Example:
            >>> loader = ActLoad(some_dir)
            >>> # If loader has keys "my.activation.first", "my.activation.second", "other.key"
            >>> filtered = loader.filter_by_prefix("my.activation.")
            >>> filtered["first"]  # Accesses "my.activation.first"
            >>> filtered["second"]  # Accesses "my.activation.second"
            >>> # Can be chained:
            >>> double_filtered = loader.filter_by_prefix("my.").filter_by_prefix("activation.")
            >>> double_filtered["first"]  # Also accesses "my.activation.first"
        """
        filtered_activations = {}
        for name, activation in self.activations.items():
            if name.startswith(prefix):
                # Strip the prefix from the key
                stripped_name = name[len(prefix) :]
                filtered_activations[stripped_name] = activation

        new_prefix_chain = self._prefix_chain + [prefix]
        return ActLoad(
            _base_activations=filtered_activations,
            _prefix_chain=new_prefix_chain,
            dir=self._dir,
        )


ActivationLoader = ActLoad



from . import actsave as actsave
from . import lovely as lovely
from . import typecheck as typecheck
from .actsave import ActLoad as ActLoad
from .actsave import ActSave as ActSave
from .collections import apply_to_collection as apply_to_collection
from .display import display as display
from .logging import init_python_logging as init_python_logging
from .logging import setup_logging as setup_logging
from .lovely import lovely_monkey_patch as lovely_monkey_patch
from .snoop import snoop as snoop
from .typecheck import tassert as tassert
from .typecheck import typecheck_modules as typecheck_modules
from .typecheck import typecheck_this_module as typecheck_this_module



from __future__ import annotations

import sys
from collections import defaultdict
from collections.abc import Callable, Sequence
from typing import Any

import numpy as np
from typing_extensions import TypedDict

from .config import LovelyConfig


class LovelyStats(TypedDict, total=False):
    """Statistics for tensor representation."""

    # Basic tensor information
    shape: Sequence[int]
    size: int
    type_name: str
    dtype_str: str
    device: str | None
    nbytes: int
    requires_grad: bool
    is_meta: bool

    # Content flags
    all_zeros: bool
    has_nan: bool
    has_pos_inf: bool
    has_neg_inf: bool
    is_complex: bool

    # Numeric statistics
    min: float | None
    max: float | None
    mean: float | None
    std: float | None

    # Complex number statistics
    mag_min: float | None
    mag_max: float | None
    real_min: float | None
    real_max: float | None
    imag_min: float | None
    imag_max: float | None

    # Representation
    values_str: str | None


# Formatting utilities
def sci_mode(f: float) -> bool:
    """Determine if a float should be displayed in scientific notation."""
    config = LovelyConfig.instance()
    return (abs(f) < 10**config.threshold_min) or (abs(f) > 10**config.threshold_max)


def pretty_str(x: Any) -> str:
    """Format a number or array for pretty display.

    Works with scalars, numpy arrays, torch tensors, and jax arrays.
    """
    if isinstance(x, int):
        return f"{x}"
    elif isinstance(x, float):
        if x == 0.0:
            return "0."

        sci = (
            sci_mode(x)
            if LovelyConfig.instance().sci_mode is None
            else LovelyConfig.instance().sci_mode
        )
        fmt = f"{{:.{LovelyConfig.instance().precision}{'e' if sci else 'f'}}}"
        return fmt.format(x)
    elif isinstance(x, complex):
        # Handle complex numbers
        real_part = pretty_str(x.real)
        imag_part = pretty_str(abs(x.imag))
        sign = "+" if x.imag >= 0 else "-"
        return f"{real_part}{sign}{imag_part}j"

    # Handle array-like objects
    try:
        if hasattr(x, "ndim") and x.ndim == 0:
            return pretty_str(x.item())
        elif hasattr(x, "shape") and len(x.shape) > 0:
            slices = [pretty_str(x[i]) for i in range(min(x.shape[0], 10))]
            if x.shape[0] > 10:
                slices.append("...")
            return "[" + ", ".join(slices) + "]"
    except:
        pass

    # Fallback
    return str(x)


def sparse_join(items: list[str | None], sep: str = " ") -> str:
    """Join non-empty strings with a separator."""
    return sep.join([item for item in items if item])


def ansi_color(s: str, col: str, use_color: bool = True) -> str:
    """Add ANSI color to a string if use_color is True."""
    if not use_color:
        return s

    style = defaultdict(str)
    style["grey"] = "\x1b[38;2;127;127;127m"
    style["red"] = "\x1b[31m"
    end_style = "\x1b[0m"

    return style[col] + s + end_style


def bytes_to_human(num_bytes: int) -> str:
    """Convert bytes to a human-readable string (b, Kb, Mb, Gb)."""
    units = ["b", "Kb", "Mb", "Gb"]

    value = num_bytes
    matched_unit: str | None = None
    for unit in units:
        if value < 1024 / 10:
            matched_unit = unit
            break
        value /= 1024.0

    assert matched_unit is not None, "No matching unit found"

    if value % 1 == 0 or value >= 10:
        return f"{round(value)}{matched_unit}"
    else:
        return f"{value:.1f}{matched_unit}"


def in_debugger() -> bool:
    """Returns True if running in a debugger."""
    return getattr(sys, "gettrace", None) is not None and sys.gettrace() is not None


# Common tensor representation
def format_tensor_stats(tensor_stats: LovelyStats, color: bool | None = None) -> str:
    """Format tensor stats into a pretty string representation."""
    conf = LovelyConfig.instance()
    if color is None:
        color = conf.color
    if in_debugger():
        color = False

    # Basic tensor info
    shape_str = str(list(shape)) if (shape := tensor_stats.get("shape")) else None
    type_str = sparse_join([tensor_stats.get("type_name"), shape_str], sep="")

    # Calculate memory usage
    numel = None
    if (size := tensor_stats.get("size")) and (nbytes := tensor_stats.get("nbytes")):
        shape = tensor_stats.get("shape", [])

        if shape and max(shape) != size:
            numel = f"n={size}"
            if conf.show_mem_above <= nbytes:
                numel = sparse_join([numel, f"({bytes_to_human(nbytes)})"])
        elif conf.show_mem_above <= nbytes:
            numel = bytes_to_human(nbytes)

    # Handle empty tensors
    if tensor_stats.get("size", 0) == 0:
        common = ansi_color("empty", "grey", color)
    # Handle all zeros
    elif tensor_stats.get("all_zeros"):
        common = ansi_color("all_zeros", "grey", color)
    # Handle complex tensors
    elif tensor_stats.get("is_complex"):
        complex_info = []

        # For magnitude stats
        if (mag_min := tensor_stats.get("mag_min")) is not None and (
            mag_max := tensor_stats.get("mag_max")
        ) is not None:
            complex_info.append(f"|z|∈[{pretty_str(mag_min)}, {pretty_str(mag_max)}]")

        # For real part stats
        if (real_min := tensor_stats.get("real_min")) is not None and (
            real_max := tensor_stats.get("real_max")
        ) is not None:
            complex_info.append(f"Re∈[{pretty_str(real_min)}, {pretty_str(real_max)}]")

        # For imaginary part stats
        if (imag_min := tensor_stats.get("imag_min")) is not None and (
            imag_max := tensor_stats.get("imag_max")
        ) is not None:
            complex_info.append(f"Im∈[{pretty_str(imag_min)}, {pretty_str(imag_max)}]")

        common = sparse_join(complex_info)
    # Handle normal tensors with stats
    elif (min_val := tensor_stats.get("min")) is not None and (
        max_val := tensor_stats.get("max")
    ) is not None:
        minmax = None
        meanstd = None

        if tensor_stats.get("size", 0) > 2:
            minmax = f"x∈[{pretty_str(min_val)}, {pretty_str(max_val)}]"

        if (
            (mean := tensor_stats.get("mean")) is not None
            and (std := tensor_stats.get("std")) is not None
            and tensor_stats.get("size", 0) >= 2
        ):
            meanstd = f"μ={pretty_str(mean)} σ={pretty_str(std)}"

        common = sparse_join([minmax, meanstd])
    else:
        common = None

    # Handle warnings
    warnings = []
    if tensor_stats.get("has_nan"):
        warnings.append(ansi_color("NaN!", "red", color))
    if tensor_stats.get("has_pos_inf"):
        warnings.append(ansi_color("+Inf!", "red", color))
    if tensor_stats.get("has_neg_inf"):
        warnings.append(ansi_color("-Inf!", "red", color))

    attention = sparse_join(warnings)
    common = sparse_join([common, attention])

    # Other tensor attributes
    dtype = tensor_stats.get("dtype_str", "")
    device = tensor_stats.get("device")
    grad = "grad" if tensor_stats.get("requires_grad") else None

    # Format values for small tensors
    vals = None
    if (
        0 < tensor_stats.get("size", 0) <= 10
        and tensor_stats.get("is_meta", False) is False
    ):
        vals = tensor_stats.get("values_str")

    # Join all parts
    result = sparse_join([type_str, dtype, numel, common, grad, device, vals])

    return result


def real_stats(array: np.ndarray) -> LovelyStats:
    stats: LovelyStats = {}

    # Check for special values
    stats["has_nan"] = bool(np.isnan(array).any())
    stats["has_pos_inf"] = bool(np.isposinf(array).any())
    stats["has_neg_inf"] = bool(np.isneginf(array).any())

    # Only compute min/max/mean/std for good data
    good_data = array[np.isfinite(array)]

    if len(good_data) > 0:
        stats["min"] = float(good_data.min())
        stats["max"] = float(good_data.max())
        stats["all_zeros"] = stats["min"] == 0 and stats["max"] == 0 and array.size > 1

        if len(good_data) > 1:
            stats["mean"] = float(good_data.mean())
            stats["std"] = float(good_data.std())

    # Get string representation of values for small tensors
    if 0 < array.size <= 10:
        stats["values_str"] = pretty_str(array)

    return stats


def complex_stats(array: np.ndarray) -> LovelyStats:
    stats: LovelyStats = {}

    # Calculate magnitude (absolute value)
    magnitude = np.abs(array)

    # Check for special values in real or imaginary parts
    stats["has_nan"] = bool(np.isnan(array.real).any() or np.isnan(array.imag).any())

    # Get statistics for magnitude
    good_mag = magnitude[np.isfinite(magnitude)]
    if len(good_mag) > 0:
        stats["mag_min"] = float(good_mag.min())
        stats["mag_max"] = float(good_mag.max())
        stats["all_zeros"] = (
            stats["mag_min"] == 0 and stats["mag_max"] == 0 and array.size > 1
        )

    # Get statistics for real and imaginary parts
    real_part = array.real
    imag_part = array.imag

    good_real = real_part[np.isfinite(real_part)]
    good_imag = imag_part[np.isfinite(imag_part)]

    if len(good_real := real_part[np.isfinite(real_part)]):
        stats["real_min"] = float(good_real.min())
        stats["real_max"] = float(good_real.max())

    if len(good_imag := imag_part[np.isfinite(imag_part)]):
        stats["imag_min"] = float(good_imag.min())
        stats["imag_max"] = float(good_imag.max())

    # Get string representation of values for small tensors
    if 0 < array.size <= 10:
        stats["values_str"] = pretty_str(array)

    return stats


def array_stats(array: np.ndarray, ignore_empty: bool = True) -> LovelyStats:
    """Compute all statistics for a given array.

    Args:
        array (np.ndarray): The input array.
        ignore_empty (bool): If True, ignore empty arrays.

    Returns:
        LovelyStats: A dictionary containing the computed statistics.
    """
    if ignore_empty and array.size == 0:
        return {}

    if np.iscomplexobj(array):
        return complex_stats(array)
    else:
        return real_stats(array)


def patch_to(
    cls: type,
    name: str,
    func: Callable[[Any], Any],
    as_property: bool = False,
) -> None:
    """Simple patch_to implementation to avoid fastcore dependency."""
    if as_property:
        setattr(cls, name, property(func))
    else:
        setattr(cls, name, func)



from __future__ import annotations

import functools
import importlib
import importlib.util
import logging
from typing import Any, Literal, TypeAlias

from typing_extensions import override

log = logging.getLogger(__name__)


class _Backend:
    def __init__(self, name: str, import_path: str, pytree_def_import: str):
        self.name = name
        self.import_path = import_path
        self.pytree_def_import = pytree_def_import

    def is_available(self) -> bool:
        """Check if the backend is available."""
        try:
            is_available = importlib.util.find_spec(self.import_path) is not None
        except ImportError:
            is_available = False

        if not is_available:
            log.warning(f"Backend '{self.name}' is not available.")
        return is_available

    @functools.cache
    def _imported_module(self):
        return importlib.import_module(self.import_path)

    def __call__(self, fn_name: str, /, *args, **kwargs):
        module = self._imported_module()
        if (fn := getattr(module, fn_name, None)) is None:
            raise AttributeError(
                f"Module '{self.import_path}' does not implement function '{fn_name}'"
            )
        return fn(*args, **kwargs)

    def pytree_def(self):
        """Return the PyTreeDef class for this backend."""
        # Split the import path to get the class name.
        module_path, cls_name = self.pytree_def_import.rsplit(".", 1)
        module = importlib.import_module(module_path)
        if (pytree_def := getattr(module, cls_name, None)) is None:
            raise AttributeError(
                f"Module '{module_path}' does not implement class '{cls_name}'"
            )
        return pytree_def

    @override
    def __repr__(self) -> str:
        return f"_Backend(name={self.name}, import_path={self.import_path})"

    @override
    def __str__(self) -> str:
        return f"{self.name} ('{self.import_path}')"


_selected_backend: _Backend | None = None

BackendName: TypeAlias = Literal["jax", "torch", "optree"]
DEFAULT_BACKENDS: dict[BackendName, _Backend] = {
    "jax": _Backend("jax", "jax.tree_util", "jax.tree_util.PyTreeDef"),
    "torch": _Backend("torch", "torch.utils._pytree", "torch.utils._pytree.TreeSpec"),
    "optree": _Backend("optree", "optree", "optree.PyTreeSpec"),
}


def set_pytree_backend(name: BackendName):
    """Set the backend for pytree operations."""
    # If not valid backend is provided, raise an error.
    if (backend := DEFAULT_BACKENDS.get(name)) is None:
        raise ValueError(
            f"Invalid backend name '{name}'. Valid options are: {', '.join(DEFAULT_BACKENDS.keys())}."
        )

    global _selected_backend
    if not backend.is_available():
        raise ImportError(f"Backend {backend} is not available.")

    _selected_backend = backend
    log.critical(f"Pytree backend set to '{backend}'.")


@functools.lru_cache(maxsize=1)
def _resolve_default_backend() -> _Backend | None:
    if (
        first_backend := next(
            (b for b in DEFAULT_BACKENDS.values() if b.is_available()), None
        )
    ) is None:
        log.warning(
            "No available backend found in defaults. "
            "Please set a backend explicitly using `nshutils.typecheck.set_pytree_backend`."
        )
        return None
    log.warning(
        f"Using backend '{first_backend.import_path}' from defaults as no explicit backend is set. "
        "Consider setting a backend explicitly using `nshutils.typecheck.set_pytree_backend`."
    )
    return first_backend


def _resolve_backend() -> _Backend:
    global _selected_backend
    # If a backend has been explicitly set, use it.
    if _selected_backend is not None:
        return _selected_backend

    # Otherwise, try to find the first available backend from the defaults,
    # however, we should strictly emit a warning if we do so.
    if (available_backend := _resolve_default_backend()) is not None:
        return available_backend

    raise RuntimeError("No available backend found.")


def _call(fn_name: str, *args: Any, **kwargs: Any) -> Any:
    backend = _resolve_backend()
    return backend(fn_name, *args, **kwargs)


def tree_flatten(*args: Any, **kwargs: Any):
    return _call("tree_flatten", *args, **kwargs)


def tree_unflatten(*args: Any, **kwargs: Any):
    return _call("tree_unflatten", *args, **kwargs)


def tree_map(*args: Any, **kwargs: Any):
    return _call("tree_map", *args, **kwargs)


def tree_structure(*args: Any, **kwargs: Any):
    return _call("tree_structure", *args, **kwargs)


def tree_leaves(*args: Any, **kwargs: Any):
    return _call("tree_leaves", *args, **kwargs)


def PyTreeDef_cls():
    """
    Return the PyTreeDef class for the currently selected backend.
    """
    backend = _resolve_backend()
    return backend.pytree_def()



from __future__ import annotations

from ._monkey_patch_all import monkey_patch as monkey_patch
from .config import LovelyConfig as LovelyConfig
from .jax_ import jax_monkey_patch as jax_monkey_patch
from .jax_ import jax_repr as jax_repr
from .numpy_ import numpy_monkey_patch as numpy_monkey_patch
from .numpy_ import numpy_repr as numpy_repr
from .torch_ import torch_monkey_patch as torch_monkey_patch
from .torch_ import torch_repr as torch_repr

lovely_monkey_patch = monkey_patch
